use std::mem;
use std::pin::Pin;
use std::task::{Context, Poll, Waker};
use std::sync::{Arc, Mutex, MutexGuard};

use futures::prelude::*;
use futures::ready;
use futures_signals::signal::{Mutable, Signal, MutableSignalCloned, MutableLockMut};

use termion::event::Key;

use crate::input::Inputs;
use crate::schema::API;

use slog::{
    Logger,
    Drain,
    Level,
    Record,
    OwnedKVList,
};

use tui::text::{Span, Spans};
use tui::style::{Style, Color};

#[derive(PartialEq, Eq, Debug, Clone)]
pub enum Window {
    Main,
    Help,
}

/// Application state struct
pub struct Sute<'a, S> {

    // TODO: BE SMART. Inputs change the state, resize signals change the state, futures completing
    // change the state.

    state: Mutable<SuteState>,

    signal: S,
    inputs: Inputs,
    api: Option<API>,
    log: Logger,

    drain: Arc<LogDrain<'a>>,
}

impl<'a, S: Unpin> Sute<'a, S> {
    pub fn new(signal: S, log: Logger, drain: Arc<LogDrain<'a>>, api: Option<API>) -> Self {
        let inputs = Inputs::new();
        let state = Mutable::new(SuteState::new());

        Self { state, signal, inputs, api, log, drain, }
    }

    fn run_cmd(&mut self, cmdline: String) {
        let mut words = cmdline.split_ascii_whitespace();
        match words.next() {
            Some("quit") => self.state.lock_mut().running = false,
            Some("connect") => {
                self.connect(cmdline)
            },
            cmd => info!(self.log, "Issues unknown cmd: {:?}", cmd),
        }
    }

    fn connect(&mut self, params: String) {
        info!(self.log, "Called connect with {}", params);
    }

    fn handle_resize(&mut self, new_size: (u16,u16)) {
        trace!(self.log, "Locking in handle_resize");
        self.state.lock_mut().size = new_size;
    }

    pub fn get_state(&self) -> SuteState {
        self.state.get_cloned()
    }

    // We can make this more efficient by not cloning. But who cares?
    pub fn signal(&self) -> impl Signal<Item=SuteState> {
        self.state.signal_cloned()
    }

    pub fn handle_key(&mut self, key: Key) {
        trace!(self.log, "Locking in handle_key");
        let mut state = self.state.lock_mut();
        match key {
            Key::Char('\n') => {
                let cmd = mem::replace(&mut state.cmd_line, String::new());
                // drop the mutably borrowed state here so we can mutably re-borrow self afterwards
                mem::drop(state);
                self.run_cmd(cmd);
            },
            Key::Char(c) => {
                state.cmd_line.push(c);
            },
            Key::Backspace => {
                state.cmd_line.pop();
            },
            _ => {
            }
        }
    }
}

impl <'a, S: Unpin + Signal<Item=(u16,u16)>> Future for Sute<'a, S> {
    type Output = ();
    fn poll(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
        // Return early if we're not running anymore
        if ! self.state.lock_ref().running {
            return Poll::Ready(());
        }

        if let Poll::Ready(Some(size)) = Pin::new(&mut self.signal).poll_change(cx) {
            self.handle_resize(size);
        }
        match ready!(Pin::new(&mut self.inputs).poll_next(cx)) {
            Some(key) => {
                self.handle_key(key);
            },
            // If the input closes stop the program
            None => {
                trace!(self.log, "Locking in impl Future for Sute");
                self.state.lock_mut().running = false;
                return Poll::Ready(());
            },
        }

        Poll::Pending
    }
}

#[derive(Debug, Clone)]
pub struct SuteState {
    pub active_win: Window,
    pub size: (u16,u16),
    pub running: bool,
    pub tick: usize,
    pub server: Option<String>,
    pub cmd_line: String,
    pub tick_c: char,

    // TODO: Figure out how to put log lines here signaled
}

impl SuteState {
    pub fn new() -> Self {
        Self {
            active_win: Window::Main,
            size: (80,20),
            running: true,
            tick: 0,
            server: None,
            cmd_line: String::new(),
            tick_c: '|',
        }
    }
}

#[derive(Debug)]
pub struct LogDrain<'a> {
    inner: Mutex<Vec<Spans<'a>>>,
}

impl<'a> LogDrain<'a> {
    pub fn new() -> Self {
        Self { inner: Mutex::new(Vec::new()) }
    }
    pub fn get_inner(&self) -> MutexGuard<Vec<Spans<'a>>> {
        self.inner.lock().unwrap()
    }
}
impl<'a> Drain for LogDrain<'a> {
    type Ok = ();
    type Err = ();
    fn log(&self, record: &Record, values: &OwnedKVList) -> Result<Self::Ok, Self::Err> {
        let critical_style: Style = Style::default().fg(Color::Magenta);
        let error_style: Style = Style::default().fg(Color::Red);
        let warning_style: Style = Style::default().fg(Color::Yellow);
        let info_style: Style = Style::default().fg(Color::Cyan);
        let debug_style: Style = Style::default().fg(Color::White);
        let trace_style: Style = Style::default().fg(Color::Green);

        let lvlspan = match record.level() {
            Level::Critical => Span::styled("CRIT", critical_style),
            Level::Error => Span::styled("ERRO", error_style),
            Level::Warning => Span::styled("WARN", warning_style),
            Level::Info => Span::styled("INFO", info_style),
            Level::Debug => Span::styled("DEBG", debug_style),
            Level::Trace => Span::styled("TRCE", trace_style),
        };
        let contentspan = Span::from(record.msg().to_string());

        {
            let mut v = self.inner.lock().unwrap();
            v.push(Spans::from(vec![
                Span::from("["), lvlspan, Span::from("]: "),
                contentspan,
            ]));
        }

        Ok(())
    }
}

